"""Creates anaglyphs and other simple stereo-related products.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging
from pathlib import Path

import numpy as np
from numpy.typing import NDArray
from PIL import Image
from skimage.io import imread
from skimage.registration import phase_cross_correlation

from vipersci import util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-a", "--anaglyph", action="store_true", help="Creates a red/cyan anaglyph"
    )
    group.add_argument(
        "-s",
        "--side_by_side",
        action="store_true",
        help="Creates a side-by-side image.",
    )
    group.add_argument(
        "-w", "--wiggle", action="store_true", help="Creates a wiggle GIF."
    )
    parser.add_argument(
        "-r",
        "--register",
        action="store_true",
        help="Performs an automatic registration which may align the images better.",
    )
    parser.add_argument(
        "left",
        type=Path,
        help="Path to image file which will be the stereo-left image.",
    )
    parser.add_argument(
        "right",
        type=Path,
        help="Path to image file which will be the stereo-right image.",
    )
    parser.add_argument(
        "out",
        type=Path,
        help="Path to write the output image to.",
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    left = imread(args.left)
    right = imread(args.right)

    if args.register:
        right = correlate_and_shift(left, right)

    if args.anaglyph:
        a = anaglyph(left, right)
        im = Image.fromarray(a)
        im.save(args.out)
    elif args.side_by_side:
        im = Image.fromarray(np.hstack((left, right)))
        im.save(args.out)
    elif args.wiggle:
        left_im = Image.fromarray(left)
        right_im = Image.fromarray(right)
        left_im.save(
            args.out,
            save_all=True,
            append_images=[
                right_im,
            ],
            duration=400,
            loop=0,  # forever
            dispose=0,
        )
    else:
        parser.error("Argument added but not handled by the program.")

    return


def anaglyph(left: NDArray, right: NDArray, align=False) -> NDArray:
    """Return a 3D array arranged as three 2D arrays derived from *left* and *right*
    that represent a red/cyan anaglyph.

    If *align* is true, then scikit-image's phase_cross_correlation function will
    be used to improve alignment between *left* and *right* which might result
    in a better anaglyph.
    """
    if left.shape != right.shape:
        raise ValueError(
            f"The left image shape {left.shape} does not match the right {right.shape}"
        )

    if align:
        right = correlate_and_shift(left, right)

    if len(left.shape) == 2:
        red = left
    elif len(left.shape) == 3 and left.shape[2] == 3:
        red = left[..., 0]
    else:
        raise ValueError(
            f"Don't know how to deal with {left.shape} dimensions in the left image."
        )

    if len(right.shape) == 2:
        green = right
        blue = right
    elif len(right.shape) == 3 and left.shape[2] == 3:
        green = right[..., 1]
        blue = right[..., 2]
    else:
        raise ValueError(
            f"Don't know how to deal with {right.shape} dimensions in the right image."
        )

    return np.dstack((red, green, blue))


def correlate_and_shift(left: NDArray, right: NDArray) -> NDArray:
    """
    Returns an array that is the result of aligning the *right* array to the
    *left* array.

    This is a thin wrapper for scikit-image's phase_cross_correlation()
    function with performs the correlation, and then shifts the array.
    """
    shift = tuple(
        phase_cross_correlation(left, right, normalization=None)[0].astype(int)
    )
    logger.info(f"Right image shift: {shift}")
    return np.roll(right, shift, (0, 1))


def create(left: Path, right: Path, output: Path, align=False):
    """
    This is a wrapper for the anaglyph() function which can be given
    Path objects to read from and write to.
    """
    left_arr = imread(left)
    right_arr = imread(right)
    a = anaglyph(left_arr, right_arr, align)
    im = Image.fromarray(a)
    im.save(output)
    return



#!/usr/bin/env python
"""This module contains a derived date and time classes which with extended
ISO 8601 handling to deal with trailing "Z" characters instead of +0.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import datetime


def fromisozformat(date_string):
    """
    Return a datetime corresponding to *date_string* in one of the formats emitted by
    isozformat().  This datetime will be a timezone-aware datetime.
    """
    try:
        dt = datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        dt = datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%fZ")

    return dt.replace(tzinfo=datetime.timezone.utc)


def isozformat(date_time, sep="T", timespec="auto"):
    """
    Return a string representing the UTC date and time in ISO 8601 format with
    the trailing letter 'Z' representing the UTC offset for the provided
    UTC *date_time* object.

    This function is similar to the standard library's
    datetime.datetime.isoformat() function (see that documentation
    for details on *sep* and *timespec*).

    The difference is that the provided *date_time* must be timezone
    aware and it must be in UTC (its utcoffset() must not be None and
    must equal zero). Otherwise, it will raise a ValueError.

    The most important difference is that rather than returning a string
    representation that ends in "+00:00" for a UTC datetime object, it
    will return a representation that ends in "Z".  Both formats are
    valid ISO 8601 date formats, but the standard library picks one
    string representation, and this function provides the other, which
    is the required format for PDS4 labels.
    """
    if date_time.utcoffset() == datetime.timedelta():
        return date_time.replace(tzinfo=None).isoformat(sep, timespec) + "Z"
    else:
        raise ValueError(
            "The datetime object is either naive (not timezone aware), "
            "or has a non-zero offset from UTC.  Maybe you just want "
            "the datetime object's isoformat() function?"
        )



#!/usr/bin/env python
"""This module has tests for the vis.pano-check functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from datetime import datetime, timezone
from functools import partial
from pathlib import Path
import unittest
from unittest.mock import patch

import pandas as pd

from vipersci.pds import pid as pds
from vipersci.vis import pano_check as pc
from vipersci.vis.db.image_records import ImageRecord


class TestFunctions(unittest.TestCase):
    def test_arg_parser(self):
        p = pc.arg_parser()
        self.assertIsInstance(p, ArgumentParser)

    def test_groupby_2nd(self):
        tuples = [
            ("Alice", "foo"),
            ("Bob", "bar"),
            ("Catherine", "bar"),
        ]
        truth = [
            (
                [
                    "Alice",
                ],
                "foo",
            ),
            (["Bob", "Catherine"], "bar"),
        ]

        grouped = pc.groupby_2nd(tuples)
        self.assertEqual(truth, grouped)


class TestPositionAndPose(unittest.TestCase):
    def setUp(self):
        self.datetimes = (
            datetime(2024, 11, 27, 1, 2, 3, tzinfo=timezone.utc),
            datetime(2024, 11, 28, 1, 2, 3, tzinfo=timezone.utc),
            datetime(2024, 11, 29, 1, 2, 3, tzinfo=timezone.utc),
            datetime(2024, 11, 30, 1, 2, 3, tzinfo=timezone.utc),
        )
        self.times = [x.timestamp() for x in self.datetimes]
        self.x = (-1, 2, 3, 4)
        self.y = (-1, 2, 3, 4)
        self.yaw = (0, 45, 90, 270)

        self.truth = [
            (t, (x, y, yaw))
            for t, x, y, yaw in zip(self.times, self.x, self.y, self.yaw)
        ]

        self.df = pd.DataFrame(
            data={
                "c0": pd.Series(self.times),
                "c1": pd.Series(self.x),
                "c2": pd.Series(self.y),
                "c3": pd.Series(self.yaw),
            }
        )

    def test_get_pp_from_mapserver(self):
        class req_dict(dict):
            def json(self):
                return self

        time_d = {}
        for t, x, y, yaw in zip(self.times, self.x, self.y, self.yaw):
            time_d[t] = req_dict(location=(x, y), yaw=yaw)

        def requests_get_se(url, params):
            return time_d[params["event_time"]]

        with patch("vipersci.vis.pano_check.requests.get", side_effect=requests_get_se):
            tpp = pc.get_position_and_pose_from_mapserver(self.times, url="foo")
            self.assertEqual(tpp, self.truth)

    def test_get_pp_from_csv(self):
        my_times = self.times.copy()
        my_times[0] += 10

        my_truth = [
            (t, (x, y, yaw)) for t, x, y, yaw in zip(my_times, self.x, self.y, self.yaw)
        ]

        with patch("vipersci.vis.pano_check.pd.read_csv", return_value=self.df):
            tpp = pc.get_position_and_pose_from_csv(my_times, Path("dummy.csv"))
            self.assertEqual(tpp, my_truth)

    def test_check(self):
        pid_list = [
            "241127-010203-ncl-a",
            "241127-010303-ncl-a",
            "241127-010403-ncl-a",
            "241127-010403-ncr-s",
            "241128-010203-ncl-a",
            "241130-010203-ncl-a",
            "241130-010303-ncl-a",
        ]
        pds.VISID(pid_list[0])
        pids = list(map(pds.VISID, pid_list))

        self.assertRaises(ValueError, pc.check, pids)

        im_recs = list()
        for x in pids:
            im_recs.append(
                ImageRecord(
                    product_id=str(x), start_time=x.datetime(), exposure_duration=111
                )
            )

        # Only two groups in the above pid_list
        truth = [
            (pids[0:3], (self.x[0], self.y[0], self.yaw[0])),
            (pids[-2:], (self.x[3], self.y[3], self.yaw[3])),
        ]

        self.assertRaises(TypeError, pc.check, pid_list, "dummy_function")

        with patch("vipersci.vis.pano_check.pd.read_csv", return_value=self.df):
            gffp = partial(pc.get_position_and_pose_from_csv, path="dummy.csv")
            p_groups = pc.check(pids, gffp)
            self.assertEqual(p_groups, truth)

            im_groups = pc.check(im_recs, gffp)
            pid_groups = list()
            for im_list, pose in im_groups:
                pid_list = list(map(pds.VISID, [x.product_id for x in im_list]))
                pid_groups.append((pid_list, pose))
            self.assertEqual(pid_groups, truth)



"""Gather Rover locations from a REST-based service.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import csv
from collections.abc import Sequence
from datetime import datetime, timezone
import getpass
import logging
from pathlib import Path
import sys

import http.client as http_client

import pandas as pd
import requests
from requests.auth import HTTPBasicAuth

from vipersci import util

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-s",
        "--start",
        default=datetime(2020, 1, 1, tzinfo=timezone.utc),
        help="An ISO8601 datetime to start the query at.",
    )
    parser.add_argument(
        "-e",
        "--end",
        default=datetime.now(tz=timezone.utc),
        help="An ISO8601 datetime to end the query at.",
    )
    parser.add_argument(
        "-f",
        "--frequency",
        help="A frequency between the start and end times, using pandas 'Offset alias' "
        "string notation (https://pandas.pydata.org/pandas-docs/stable/user_guide/"
        "timeseries.html#offset-aliases). Could be 'S' for secondly.",
    )
    parser.add_argument(
        "-u",
        "--url",
        required=True,
        help="URL to which event_times may be posted and for which location and "
        "orientation will be returned.",
    )
    parser.add_argument("--user", help="Username to authenticate to URL with.")
    parser.add_argument(
        "-p",
        "--password",
        action="store_true",
        help="If given will trigger asking for a password.",
    )
    parser.add_argument(
        "-o", "--output", type=Path, help="Output path for CSV file output."
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    if args.verbose >= 2:
        http_client.HTTPConnection.debuglevel = 1

    basic_auth = None
    if args.password:
        if args.user:
            username = args.user
        else:
            username = getpass.getuser()

        print(f"For username {username}")
        basic_auth = HTTPBasicAuth(username, getpass.getpass())

    t_start = datetime.fromisoformat(args.start)
    t_end = datetime.fromisoformat(args.end)

    if args.frequency:
        times = pd.date_range(t_start, t_end, freq=args.frequency).tolist()

        unix_times = []
        for t in times:
            unix_times.append(t.timestamp())

        tpp = get_position_and_pose(unix_times, args.url, auth=basic_auth)
    else:
        tpp = get_position_and_pose_range(
            t_start.timestamp(), t_end.timestamp(), args.url, auth=basic_auth
        )

    if args.output is not None:
        with open(args.output, "w", newline="") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(["UTC datetime", "x", "y", "yaw"])
            for row in tpp:
                dt = datetime.fromtimestamp(row[0], tz=timezone.utc)
                writer.writerow(
                    [
                        dt.isoformat(),
                    ]
                    + list(row[1:])
                )


def get_position_and_pose(times: list, url: str, crs: str = "VIPER:910101", auth=None):
    """
    Given a list of unix times and a URL that requests can be made against,
    return a list of two-tuples whose first element is the time and
    whose second element is a three-tuple of x-location, y-location,
    and yaw.
    """

    tpp = list()

    for t in times:
        logger.info(f"unix timestamp: {t}")
        position_result = requests.get(
            url,
            params={
                "event_time": t,  # Event time in unix datetime.
                "margin_seconds": 2,  # Some "margin time" around the event to search.
                "format": "xyyaw",  # Could also be xyyaw_uncertainty
                "time_format": "unix_seconds",
                "source": "ROVER",
                "limit": 0,
                "crs_code": crs,
            },
            verify=False,
            auth=auth,
        )
        logger.debug(position_result)
        rj = position_result.json()
        logger.info(rj)

        tpp.append(
            (rj["event_seconds"], rj["location"][0], rj["location"][1], rj["yaw"])
        )

    return tpp


def get_position_and_pose_range(
    start_time, stop_time, url: str, crs: str = "VIPER:910101", auth=None
):
    tpp = list()

    track_result = requests.get(
        url,
        params={
            "min_time": start_time,
            "max_time": stop_time,
            "all": False,
            "format": "xyyaw",
            "crs_code": crs,
            "time_format": "unix_seconds",
            "source": "ROVER",
            "start_end_only": False,
            "simplify": False,
            "order": "asc",
        },
        verify=False,
        auth=auth,
    )
    logger.debug(track_result)
    rj = track_result.json()
    logger.info(rj)

    if isinstance(rj["event_seconds"], Sequence):
        for time, loc, yaw in zip(rj["event_seconds"], rj["location"], rj["yaw"]):
            tpp.append((time, loc[0], loc[1], yaw))
    else:
        tpp.append(
            (rj["event_seconds"], rj["location"][0], rj["location"][1], rj["yaw"])
        )

    return tpp


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""This module has tests for functions used internally in heatmaps.heatmap"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import numpy as np
import rasterio
from rasterio.coords import BoundingBox
import shapely
import unittest

from vipersci.carto import heatmap, bounds


class TestBounds(unittest.TestCase):
    def test_simple(self):
        initial_bounds = (0, 0, 10, 10)
        expected_bounds = initial_bounds
        out_bounds = bounds.pad_grid_align_bounds(*initial_bounds, 1)
        self.assertEqual(out_bounds, expected_bounds)

    def test_padded(self):
        initial_bounds = (0, 0, 10, 10)
        expected_bounds = (-1, -1, 11, 11)
        out_bounds = bounds.pad_grid_align_bounds(*initial_bounds, 1, padding=1)
        self.assertEqual(out_bounds, expected_bounds)

    def test_small_gsd(self):
        initial_bounds = (0, 0, 10, 10)
        expected_bounds = initial_bounds
        out_bounds = bounds.pad_grid_align_bounds(*initial_bounds, 0.1)
        self.assertEqual(out_bounds, expected_bounds)

    def test_origin_adjust(self):
        initial_bounds = (1, 1, 9, 9)
        expected_bounds = (0, 0, 10, 10)
        out_bounds = bounds.pad_grid_align_bounds(*initial_bounds, 2)
        self.assertEqual(out_bounds, expected_bounds)

    def test_buffered_mask_full(self):
        points = shapely.geometry.LineString([(0, 0), (1, 0), (1, 1), (0, 1)])
        gsd = 1
        b = BoundingBox(*bounds.pad_grid_align_bounds(0, 0, 1, 1, gsd, 1))
        t = rasterio.transform.from_origin(b.left, b.top, gsd, gsd)
        mask = heatmap.buffered_mask(points, t, buffer=1)
        self.assertTrue(np.array_equal(np.full((3, 3), False), mask))



#!/usr/bin/env python
"""This module has tests for the nss module."""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from pathlib import Path
import unittest
from unittest.mock import patch

import numpy as np
from scipy.interpolate import RegularGridInterpolator

from vipersci import nss


class TestUniformWEH(unittest.TestCase):
    def test_errors(self):
        self.assertRaises(ValueError, nss.uniform_weh, 10, c0=5)
        self.assertRaises(ValueError, nss.uniform_weh, 0)

    def test_uniform_weh(self):
        self.assertAlmostEqual(0.0397245, nss.uniform_weh(10))
        np.testing.assert_allclose(
            np.array([0.03972452, 0.00900205]),
            nss.uniform_weh(np.array((10, 20))),
            rtol=1e-06,
        )


class TestModel(unittest.TestCase):
    def setUp(self) -> None:
        self.text = list(
            (
                "-99, 0, 0.5, 1",
                "60, 1, 2, 3",
                "30, 10, 20, 30",
                "10, 100, 200, 300",
            )
        )
        self.rc = np.array([10, 30, 60])
        self.cc = np.array([0, 0.5, 1])
        self.arr = np.array(
            [
                [100, 200, 300],
                [10, 20, 30],
                [1, 2, 3],
            ]
        )
        return

    def test_read(self):
        arr, row_coords, col_coords = nss.read_csv(self.text)
        np.testing.assert_array_equal(self.rc, row_coords)
        np.testing.assert_array_equal(self.cc, col_coords)
        np.testing.assert_array_equal(self.arr, arr)

    def test_model(self):
        with patch("vipersci.nss.read_csv", return_value=(self.arr, self.rc, self.cc)):
            m = nss.model(Path("dummy"))
            self.assertIsInstance(m, RegularGridInterpolator)
            np.testing.assert_array_equal(np.array(88), m((0.3, 20)))


class TestSimulator(unittest.TestCase):
    def setUp(self) -> None:
        self.rc = np.array([10, 30, 60])
        self.cc = np.array([0, 0.5, 1])
        self.arr = np.array(
            [
                [100, 200, 300],
                [10, 20, 30],
                [1, 2, 3],
            ]
        )

    def testInit(self):
        with patch("vipersci.nss.read_csv", return_value=(self.arr, self.rc, self.cc)):
            ds = nss.DataSimulator(Path("dummy1"), Path("dummy2"))
            self.assertIsInstance(ds, nss.DataSimulator)

    def testCall(self):
        with patch("vipersci.nss.read_csv", return_value=(self.arr, self.rc, self.cc)):
            ds = nss.DataSimulator(Path("dummy1"), Path("dummy2"))
            self.assertRaises(ValueError, ds, 5, 5)
            d1, d2 = ds(0.3, 20)
            self.assertEqual(88, d1)
            self.assertEqual(88, d2)

            d1_arr, d2_arr = ds([0.3, 0.5], [20, 30])
            np.testing.assert_array_almost_equal(np.array([88, 20]), d1_arr)
            np.testing.assert_array_almost_equal(np.array([88, 20]), d2_arr)


class TestModeler(unittest.TestCase):
    def setUp(self) -> None:
        self.rc = np.array([10, 30, 60])
        self.cc = np.array([0, 0.5, 1])
        self.arr = np.array(
            [
                [100, 200, 300],
                [10, 20, 30],
                [1, 2, 3],
            ]
        )

    def test_init(self):
        with patch("vipersci.nss.read_csv", return_value=(self.arr, self.rc, self.cc)):
            ds = nss.DataModeler(Path("dummy1"), Path("dummy2"))
            self.assertIsInstance(ds, nss.DataModeler)

    def test_call(self):
        with patch("vipersci.nss.read_csv", return_value=(self.arr, self.rc, self.cc)):
            m = nss.DataModeler(Path("dummy"), Path("dummy2"))
            d1 = np.ma.array([1, 2, 3], mask=[1, 0, 1])
            d2 = np.ma.array([1, 2, 3], mask=[0, 1, 1])
            self.assertRaises(ValueError, m, d1, d2)

        with patch("vipersci.nss.read_csv", return_value=(self.arr, self.rc, self.cc)):
            m = nss.DataModeler(Path("dummy"), Path("dummy2"))
            bd_arr, weh_arr, uweh_arr = m(
                [0.1, 0.2, 0.3, 0.4, 0.5], [10, 20, np.nan, 40, 50]
            )
            t = np.array(
                [
                    120.0,
                    77.0,
                    np.nan,
                    12.6,
                    8.0,
                ]
            )
            np.testing.assert_array_almost_equal(t, bd_arr)
            np.testing.assert_array_almost_equal(t, weh_arr)
            np.testing.assert_array_equal(
                np.array([np.nan, np.nan, np.nan, np.nan, np.nan]), uweh_arr
            )

        with patch("vipersci.nss.read_csv", return_value=(self.arr, self.rc, self.cc)):
            m = nss.DataModeler(Path("dummy"), Path("dummy2"))
            d1 = np.ma.array([0.1, 0.2, 0.3, 0.4, 0.5], mask=[0, 0, 0, 0, 1])
            d2 = np.ma.array([10, 20, np.nan, 40, 50], mask=[0, 0, 0, 0, 1])
            bd_arr, weh_arr, uweh_arr = m(d1, d2)
            t = np.array(
                [
                    120.0,
                    77.0,
                    np.nan,
                    12.6,
                    np.nan,
                ]
            )
            np.testing.assert_array_almost_equal(t, bd_arr)
            np.testing.assert_array_almost_equal(t, weh_arr)
            np.testing.assert_array_equal(
                np.array([np.nan, np.nan, np.nan, np.nan, np.nan]), uweh_arr
            )



"""Creates VIS Image Records.

This module builds VIS Image Records from input data.  A VIS Image Record
consists of a TIFF file and a JSON file containing meta-data.
The meta data can optionally be inserted into a database

For now, this program still has a variety of hard-coded elements,
that will eventually be extracted from telemetry.

The command-line version is primarily to aide testing.
"""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from datetime import datetime, timezone
import json
import logging
from typing import Union, Optional
from pathlib import Path

import numpy as np
import numpy.typing as npt
from PIL import Image
from PIL.PngImagePlugin import PngInfo
from skimage.io import imread, imsave  # maybe just imageio here?
from skimage.util import img_as_ubyte, img_as_uint
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from tifftools import read_tiff

import vipersci
from vipersci.vis.db.image_records import ImageRecord
from vipersci.vis.db.image_requests import ImageRequest  # noqa
from vipersci.vis.db.junc_image_pano import JuncImagePano  # noqa
from vipersci.vis.db.junc_image_record_tags import JuncImageRecordTag  # noqa
from vipersci.vis.db.junc_image_req_ldst import JuncImageRequestLDST  # noqa
from vipersci.vis.db.pano_records import PanoRecord  # noqa
from vipersci.pds import pid as pds
from vipersci import util

logger = logging.getLogger(__name__)

ImageType = Union[npt.NDArray[np.uint16], npt.NDArray[np.uint8]]


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-d",
        "--dburl",
        help="Database with a raw_products table which will be written to. "
        "If not given, no database will be written to.  Example: "
        "postgresql://postgres:NotTheDefault@localhost/visdb",
    )
    parser.add_argument(
        "--nojson",
        action="store_false",
        dest="json",
        help="Disables creation of .json output, which is the default.",
    )
    parser.add_argument(
        "--image",
        type=Path,
        help="Optional path to source image file which will be read and "
        "converted into a TIFF product.  If no path is given, then "
        "Only a .json file will be written and no image.",
    )
    parser.add_argument(
        "--tiff",
        type=Path,
        help="Optional pre-existing TIFF file (presumably created by this program). "
        "This file will be inspected and its information added to the output. "
        "If --image is given, this option will be ignored.",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        default=Path.cwd(),
        help="Output directory for label. Defaults to current working directory.  "
        "Output file names are fixed based on product_id, and will be over-written.",
    )
    parser.add_argument("input", type=Path, help="JSON file containing metadata.")
    return parser


def main():
    args = arg_parser().parse_args()
    util.set_logger(args.verbose)

    with open(args.input) as f:
        jsondata = json.load(f)

    if args.image is not None:
        image = imread(str(args.image))
    elif args.tiff is not None:
        image = args.tiff
    else:
        image = None

    if args.dburl is None:
        create(jsondata, image, args.output_dir, None, args.json)
    else:
        engine = create_engine(args.dburl)
        with Session(engine) as session:
            create(
                jsondata,
                image,
                args.output_dir,
                session,
                args.json,
            )
            session.commit()

    return


def create(
    metadata: dict,
    image: Union[ImageType, Path, None] = None,
    outdir: Path = Path.cwd(),
    session: Optional[Session] = None,
    json: bool = True,
    imgtype="tif",
):
    """
    Creates a TIFF file / JSON file pair in *outdir* based on the provided
    meta-data.  Returns an ImageRecord.

    If *image* is a numpy array, that array will be considered the
    Array_2D_Image and will be written to a TIFF file with the same
    naming scheme as the JSON file in *outdir*.  If *image* is a
    file path to a TIFF file, the TIFF file at that path will be
    evaluated and its meta-data added to the JSON info, as if that
    file is the file that would have been written by this function.  If
    *image* is None (the default), no TIFF file will be written,
    and certain elements of the JSON info (pertaining to the
    file) will be empty.

    If a path is provided to *outdir* the JSON and optional TIFF
    file will be written there. Defaults to the current working
    directory.

    If *session* is given, information for the image record will be
    written to the image_records table.  If not, no database activity
    will occur.
    """
    # This arrangement of creating the output files first, ensures a clean
    # db insert (otherwise the outfile-writing would need to happen in the
    # context of the session).  However, if the db insert fails, the files
    # already exist on disk.  Is that a problem?  Maybe that's fine?
    rp = make_image_record(metadata, image, outdir, imgtype)

    if json:
        write_json(rp.asdict(), outdir)

    if session is not None:
        session.add(rp)

    return rp


def check_bit_depth(pid: pds.VISID, bit_depth: Union[int, str, np.dtype]):
    """If the provided *bit_depth* is incompatible with *pid* a ValueError will be
    raised.

    The *bit_depth* will attempt to be converted to a valid integer value the bit
    depth of the PDS TIFF file.  The *bit_depth* argument can be an integer, a
    numpy dtype, and even a string compatible with the PDS4
    Array_2D_Image/Element_Array/data_type attribute.
    """

    bd = None
    if isinstance(bit_depth, int):
        bd = bit_depth
    elif isinstance(bit_depth, str):
        if bit_depth.endswith("Byte"):
            bd = 8
        else:
            bd = int(bit_depth[-1]) * 8
    elif isinstance(bit_depth, np.dtype):
        if bit_depth == np.uint16:
            bd = 16
        elif bit_depth == np.uint8:
            bd = 8

    if bd is None:
        raise ValueError(f"Cannot determine a bit-depth from {bit_depth}")

    if pid.compression == "s":
        if bd != 8:
            raise ValueError(
                f"This is a SLoG product ({pid}), but this image is not 8-bit, "
                f"it is {bit_depth}"
            )
    else:
        if bd != 16:
            raise ValueError(
                f"This product ({pid}) should be 16-bit, but it is {bit_depth}"
            )
    return


def make_image_record(
    metadata: dict,
    image: Union[ImageType, Path, None] = None,
    outdir: Path = Path.cwd(),
    imgtype="tif",
) -> ImageRecord:
    """
    Returns an ImageRecord created from the provided meta-data, and
    if *image* is a numpy array, it will write out an image file
    in TIFF format (default or if *imgtype* is "tif") or PNG format
    (if *imgtype* is "png") in *outdir* (defaults to current
    working directory).  The *imgtype* is ignored if *image* is
    a Path to an existing image file.
    """
    ir = ImageRecord(**metadata)
    pid = pds.VISID(ir.product_id)

    if image is not None:
        if isinstance(image, Path):
            if image.suffix.casefold() in (".tif", ".tiff"):
                img_d = tif_info(image)
            elif image.suffix.casefold() == ".png":
                img_d = png_info(image)
            else:
                raise ValueError(f"Don't know how to evaluate {image}")
        else:
            if imgtype.casefold() == "tif":
                img_d = tif_info(write_tiff(pid, image, outdir))
            elif imgtype.casefold() == "png":
                img_d = png_info(write_png(pid, image, outdir))
            else:
                raise ValueError(f"The value {imgtype} is not a recognized imgtype.")

        for k in ("lines", "samples"):
            if getattr(ir, k) != img_d[k]:
                raise ValueError(
                    f"The value of {k} from the image ({img_d[k]}) does not "
                    f"match the value from the metadata ({getattr(ir, k)})"
                )

        if "file_data_type" in img_d:
            check_bit_depth(pid, img_d["file_data_type"])

        ir.update(img_d)
    else:
        ir.update(
            {
                "file_byte_offset": None,
                # "file_creation_datetime": None,
                "file_data_type": None,
                "file_md5_checksum": None,
                "file_path": None,
            }
        )

    ir.update(
        {
            "software_name": "vipersci",
            "software_version": vipersci.__version__,
            "software_program_name": __name__,
        }
    )

    return ir


def file_info(p: Path) -> dict:
    """
    Returns a dict containing meta-data from a file at the provided path, *p*.
    """
    dt = datetime.fromtimestamp(p.stat().st_mtime, timezone.utc)

    d = {
        "file_creation_datetime": dt,
        "file_md5_checksum": util.md5(p),
        "file_path": p.name,
    }
    return d


def png_info(p: Path) -> dict:
    """
    Returns a dict containing meta-data from the PNG file at the
    provided path, *p*.
    """
    im = Image.open(p)

    d = {
        # "file_data_type": f"Unsigned{dt_end}",
        "lines": im.size[1],
        "samples": im.size[0],
    }
    d.update(file_info(p))
    return d


def tif_info(p: Path) -> dict:
    """
    Returns a dict containing meta-data from the TIFF file at the
    provided path, *p*.
    """
    info = read_tiff(str(p))
    tags = info["ifds"][0]["tags"]

    end = "MSB" if info["bigEndian"] else "LSB"

    # Tag 258 is bits per pixel:
    bpp = int(tags[258]["data"][0] / 8)

    if bpp == 1:
        dt_end = "Byte"
    else:
        dt_end = end + str(bpp)

    d = {
        "file_byte_offset": tags[273]["data"][0],  # Tag 273 is StripOffsets
        "file_data_type": f"Unsigned{dt_end}",
        "lines": tags[257]["data"][0],  # Tag 257 is ImageWidth,
        "samples": tags[256]["data"][0],  # Tag 256 is ImageWidth,
    }
    d.update(file_info(p))
    return d


def write_json(product: dict, outdir: Path = Path.cwd()):
    """
    Convenience function to write *product* as a JSON file in
    *outdir*.
    """
    out_path = (outdir / product["product_id"]).with_suffix(".json")
    with out_path.open("w") as f:
        json.dump(product, f, indent=2, sort_keys=True)


def write_tiff(pid: pds.VISID, image: ImageType, outdir: Path = Path.cwd()) -> Path:
    """
    Returns the path where a TIFF with a name based on *pid* and the array
    *image* was written in *outdir* (defaults to current working directory).
    """
    if image.dtype == np.bool_:
        image = img_as_ubyte(image)
    if image.dtype == np.int32:
        image = img_as_uint(image)

    check_bit_depth(pid, image.dtype)

    desc = f"VIPER {pds.vis_instruments[pid.instrument]} {pid}"

    logger.debug(desc)
    outpath = (outdir / str(pid)).with_suffix(".tif")

    imsave(str(outpath), image, check_contrast=False, description=desc, metadata=None)
    return outpath


def write_png(pid: pds.VISID, image: ImageType, outdir: Path = Path.cwd()) -> Path:
    """
    Returns the path where a PNG with a name based on *pid* and the array
    *image* was written in *outdir* (defaults to current working directory).
    """
    if image.dtype == np.bool_:
        image = img_as_ubyte(image)
    if image.dtype == np.int32:
        image = img_as_uint(image)

    check_bit_depth(pid, image.dtype)

    desc = f"VIPER {pds.vis_instruments[pid.instrument]} {pid}"

    info = PngInfo()
    info.add_itxt("Title", desc)

    logger.debug(desc)
    outpath = (outdir / str(pid)).with_suffix(".png")

    imsave(str(outpath), image, check_contrast=False, pnginfo=info)
    return outpath



"""
The dotmap module takes scalar values with 2D coordinates and creates a simple
"dot" based map with a filled circle of a given value at each input point.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import math
from typing import Tuple, Sequence

import numpy as np
from numpy.typing import NDArray
import rasterio
from rasterio.coords import BoundingBox
from rasterio.features import rasterize
from shapely.geometry import Point

from vipersci.carto.bounds import compute_bounds, pad_grid_align_bounds
from vipersci.carto.heatmap import as_ndarray


def generate_dotmap(
    x_coords: Sequence,
    y_coords: Sequence,
    values: Sequence,
    radius: float,
    ground_sample_distance: float,
    padding: int = 0,
    nodata: float = -1,
) -> Tuple[rasterio.Affine, NDArray[np.float32]]:
    """
    Creates a simple "dotmap" by drawing a filled circle with the
    provided value at each point.  Subsequent points will be drawn
    on top of any other circles they may overlap.

    Parameters:
        x_coords: x coordinates of the data points
        y_coords: y coordinates of the data points
        values: values of the data points
        ground_sample_distance: Spatial resolution of the output data
        radius: The radius of the circles to be drawn at each input point
        padding: Square padding in pixels to add to the bounds of data when
            returning an array.  If None (the default), the value of *radius*
            converted to pixels will be used.
        nodata: the fill value to use where points are not present. Defaults to -1.
    Returns:
        A tuple (transform, output)
        transform: transform used to georeference the output data
        output: the output data containing circles drawn at each point.
    """

    padding = padding + math.ceil(radius / ground_sample_distance)
    bounds = BoundingBox(
        *pad_grid_align_bounds(
            *compute_bounds(as_ndarray(x_coords), as_ndarray(y_coords)),
            ground_sample_distance,
            padding
        )
    )

    transform = rasterio.transform.from_origin(
        bounds.left, bounds.top, ground_sample_distance, ground_sample_distance
    )
    out_shape = (
        math.floor((bounds.top - bounds.bottom) / ground_sample_distance),
        math.floor((bounds.right - bounds.left) / ground_sample_distance),
    )

    shapes = []
    for x, y, val in zip(x_coords, y_coords, values):
        shapes.append((Point(x, y).buffer(radius), val))

    output = rasterize(
        shapes, out_shape=out_shape, fill=nodata, transform=transform, dtype=np.float32
    )
    return transform, output



#!/usr/bin/env python
"""This module has tests for the vis.pds.create_raw functions."""

# Copyright 2022-2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from pathlib import Path
import unittest
from unittest.mock import patch

from geoalchemy2 import load_spatialite
from sqlalchemy import create_engine
from sqlalchemy.event import listen
from sqlalchemy.orm import Session

from vipersci.vis.db import Base
from vipersci.vis.db.image_records import ImageRecord
from vipersci.vis.db.light_records import LightRecord, luminaire_names
from vipersci.vis.pds import create_raw as cr

from datetime_sqlite import isozformat


class TestParser(unittest.TestCase):
    def test_arg_parser(self):
        p = cr.arg_parser()
        self.assertIsInstance(p, ArgumentParser)
        self.assertRaises(SystemExit, p.parse_args)
        d = vars(
            p.parse_args(
                ["--dburl", "db://foo:username@host/db", "product_id_goes_here"]
            )
        )
        self.assertIn("dburl", d)
        self.assertIn("template", d)
        self.assertIn("tiff", d)
        self.assertIn("output_dir", d)
        self.assertIn("input", d)


class TestDatabase(unittest.TestCase):
    def setUp(self) -> None:
        self.ir = ImageRecord(
            adc_gain=0,
            auto_exposure=0,
            byteQuota=341,
            cameraId=0,
            capture_id=1,
            exposure_duration=111,
            file_byte_offset=256,
            file_creation_datetime="2023-08-01T00:51:51.148919Z",
            file_data_type="UnsignedLSB2",
            file_md5_checksum="b8f1a035e39c223e2b7e236846102c29",
            file_path="231125-140416-ncl-d.tif",
            imageDepth=2,
            image_id=0,
            immediateDownloadInfo=10,
            instrument_name="NavCam Left",
            instrument_temperature=0,
            light_on_hap=None,
            light_on_has=None,
            light_on_hcp=None,
            light_on_hcs=None,
            light_on_hfp=None,
            light_on_hfs=None,
            light_on_nl=None,
            light_on_nr=None,
            lines=2048,
            lobt=1700921056,
            offset=0,
            outputImageType="LOSSY_ICER_IMAGE",
            output_image_mask=8,
            padding=0,
            pga_gain=1.0,
            processing_info=10,
            product_id="231125-140416-ncl-d",
            samples=2048,
            software_name="vipersci",
            software_program_name="vipersci.vis.create_image",
            software_version="0.6.0-dev",
            start_time="2023-11-25T14:04:16Z",
            stereo=1,
            stop_time="2023-11-25T14:04:16.000111Z",
            voltage_ramp=0,
            yamcs_generation_time="2023-08-01T00:51:41.820000Z",
            yamcs_name="/ViperGround/Images/ImageData/Navcam_left_icer",
        )
        self.engine = create_engine("sqlite:///:memory:")
        listen(self.engine, "connect", load_spatialite)
        self.session = Session(self.engine)
        Base.metadata.create_all(self.engine)
        self.session.add(self.ir)
        self.session.commit()

    def tearDown(self):
        Base.metadata.drop_all(self.engine)

    def test_get_lights(self):
        truth = {k: False for k in luminaire_names.values()}
        light_name = list(luminaire_names.values())[0]

        # No lights in the db yet
        li = cr.get_lights(self.ir, self.session)
        self.assertEqual(truth, li)

        # Add an early light
        self.session.add(
            LightRecord(
                name=light_name,
                on=True,
                datetime="2023-11-25T13:04:10Z",
            )
        )
        early = cr.get_lights(self.ir, self.session)
        self.assertEqual(truth, early)

        # A previous light "off" event w/in 10 s
        self.session.add(
            LightRecord(
                name=light_name,
                on=False,
                datetime="2023-11-25T13:04:09Z",
            )
        )
        off = cr.get_lights(self.ir, self.session)
        self.assertEqual(truth, off)

        # Add a light "on" w/in 10 s:
        self.session.add(
            LightRecord(
                name=light_name,
                on=True,
                datetime="2023-11-25T14:04:10Z",
            )
        )
        self.session.add(
            LightRecord(
                name=light_name,
                on=False,
                datetime="2023-11-25T14:04:20Z",
            )
        )
        truth[light_name] = True
        t = cr.get_lights(self.ir, self.session)
        self.assertEqual(truth, t)

        # Now, no session
        t = cr.get_lights(self.ir, None)
        all_false = {k: False for k in luminaire_names.values()}
        self.assertEqual(all_false, t)

    def test_label_dict(self):
        d = cr.label_dict(self.ir, cr.get_lights(self.ir, self.session))
        self.assertEqual(
            d["lid"], f"urn:nasa:pds:viper_vis:data_raw:{self.ir.product_id}"
        )
        self.assertEqual(d["exposure_type"], "Manual")
        self.assertEqual(d["luminaires"][list(luminaire_names.values())[0]], "Off")
        self.assertEqual(d["image_filters"], "Flat field normalization. Linearization.")
        self.assertEqual(d["sample_bits"], 12)

    @patch("vipersci.vis.pds.create_raw.create_engine")
    @patch("vipersci.vis.pds.create_raw.write_xml")
    @patch("vipersci.vis.pds.create_raw.tif_info")
    def test_main(self, m_tif_info, m_write_xml, m_create_engine):
        with patch("vipersci.vis.pds.create_raw.Session", return_value=self.session):
            pa_ret_val = cr.arg_parser().parse_args(
                [
                    "--dburl",
                    "db://foo:username@host/db",
                    self.ir.product_id,
                ]
            )
            with patch("vipersci.vis.pds.create_raw.arg_parser") as parser:
                parser.return_value.parse_args.return_value = pa_ret_val
                with patch("vipersci.vis.db.image_records.isozformat", new=isozformat):
                    cr.main()
                    m_write_xml.assert_called_once()
                    (metadata, template, outdir) = m_write_xml.call_args[0]
                    self.assertEqual(metadata["product_id"], self.ir.product_id)
                    self.assertEqual(outdir, Path.cwd())
                    self.assertEqual(template, "raw-template.xml")



#!/usr/bin/env python
"""This module has tests for the nirvss module."""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import unittest

import numpy as np

from vipersci import nirvss


class TestH20(unittest.TestCase):
    def test_h2o(self):
        self.assertAlmostEqual(4.37177e-05, nirvss.band_depth_H2O(264, 2.5))
        np.testing.assert_allclose(
            np.array([4.735892e-05, 3.877421e-05]),
            nirvss.band_depth_H2O(np.array((260, 270)), np.array((2.5, 2.5))),
            rtol=1e-06,
        )


class TestOH(unittest.TestCase):
    def test_oh(self):
        self.assertAlmostEqual(0.0098995, nirvss.band_depth_OH(248))
        np.testing.assert_allclose(
            np.array([0.009899, 0.011533]),
            nirvss.band_depth_OH(np.array((248, 283))),
            rtol=5e-5,
        )



"""Creates VIS Panorama Products.

This module is meant to kick off the process to take a series of
VIS images and create a Panorama Product from them.

At this time, this module is currently just a stub, and only dumbly
mosaics the provided images together, it does not yet return a proper
Panorama Product, only a mock-up of one.

"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from datetime import timezone
import logging
from typing import Any, Dict, Union, Optional, MutableSequence, List
from pathlib import Path

import numpy as np
import numpy.typing as npt
from skimage.exposure import rescale_intensity
from skimage.io import imread, imsave  # maybe just imageio here?
from skimage.transform import resize
from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session

import vipersci
from vipersci.vis.db.image_records import ImageRecord
from vipersci.vis.db.junc_image_pano import JuncImagePano
from vipersci.vis.db.pano_records import PanoRecord
from vipersci.pds import pid as pds
from vipersci.vis.create_image import tif_info, write_json
from vipersci import util

logger = logging.getLogger(__name__)

ImageType = Union[npt.NDArray[np.uint16], npt.NDArray[np.uint8]]


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-b",
        "--bottom",
        nargs="*",
        help="Any AftCam and front-down NavCam images that should go on the 'bottom "
        "row.'  Literal '-' can be provided to indicate blank positions on the bottom "
        "row.  This is an emphemeral option that is expected to go away when we can "
        "get our hands on better pointing information.",
    )
    parser.add_argument(
        "-d",
        "--dburl",
        help="Database with a raw_products table and a panorama_products table which "
        "will be read from and written to. If not given, no database will be "
        "written to. "
        "Example: postgresql://postgres:NotTheDefault@localhost/visdb",
    )
    parser.add_argument(
        "--nojson",
        action="store_false",
        dest="json",
        help="Disables creation of .json output.",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        default=Path.cwd(),
        help="Output directory for label. Defaults to current working directory.  "
        "Output file names are fixed based on product_id, and will be over-written.",
    )
    parser.add_argument(
        "--prefix",
        type=Path,
        default=Path.cwd(),
        help="A directory path that, if given, will be prepended to paths given via "
        "inputs or will be prepended to the file_path values returned from a database "
        "query.",
    )
    parser.add_argument(
        "-t",
        "--thumb",
        nargs="?",
        default=None,  # If not given.
        const=1024,  # If just -t with no arg is given.
        help="If provided, will also generate a JPEG thumbnail, whose largest dimension"
        "will be %(const)s unless a different integer is indicated.",
    )
    parser.add_argument(
        "inputs", nargs="+", help="Either VIS raw product IDs or files."
    )
    return parser


def main():
    args = arg_parser().parse_args()
    util.set_logger(args.verbose)

    if args.dburl is None:
        create(
            args.inputs,
            args.prefix,
            args.output_dir,
            None,
            args.json,
            args.bottom,
            args.thumb,
        )
    else:
        engine = create_engine(args.dburl)
        with Session(engine) as session:
            create(
                args.inputs,
                args.prefix,
                args.output_dir,
                session,
                args.json,
                args.bottom,
                args.thumb,
            )
            session.commit()

    return


def create(
    inputs: MutableSequence[Union[Path, pds.VISID, ImageRecord, str]],
    prefixdir: Optional[Path] = None,
    outdir: Path = Path.cwd(),
    session: Optional[Session] = None,
    json: bool = True,
    bottom_row: Optional[MutableSequence[Union[Path, str]]] = None,
    thumb=None,
):
    """
    Creates a Panorama Product in *outdir*. Returns None.

    At this time, *inputs* should be a list of file paths or product IDs.

    If a path is provided to *outdir* the created files
    will be written there.

    If *session* is given, information for the PanoRecord will be
    written to the pano_records table.  If not, no database activity
    will occur.
    """

    metadata: Dict[str, Any] = dict(
        source_pids=[],
    )
    source_paths = []
    image_records = []

    for i, vid in enumerate(inputs):
        if isinstance(vid, str):
            temp_vid = pds.VISID(vid)
            if str(temp_vid) == vid:
                vid = temp_vid
            else:
                continue

        if isinstance(vid, pds.VISID):
            if session is not None:
                ir = session.scalars(
                    select(ImageRecord).where(ImageRecord.product_id == str(vid))
                ).first()
                if ir is None:
                    raise ValueError(f"{vid} was not found in the database.")
                else:
                    inputs[i] = ir

    for inp in inputs:
        if isinstance(inp, ImageRecord):
            metadata["source_pids"].append(inp.product_id)
            source_paths.append(
                inp.file_path if prefixdir is None else prefixdir / inp.file_path
            )
            image_records.append(inp)
        elif isinstance(inp, (Path, str)):
            metadata["source_pids"].append([str(pds.VISID(inp))])
            source_paths.append(inp if prefixdir is None else prefixdir / inp)
        else:
            raise ValueError(
                f"an element in input is not the right type: {inp} ({type(inp)})"
            )

    # At this time, image pointing information is not available, so we assume that
    # the images provided are provided in left-to-right order and fake these values:
    half_width = (len(inputs) / 2) * 60
    metadata["rover_pan_min"] = -1 * half_width
    metadata["rover_pan_max"] = half_width
    metadata["rover_tilt_max"] = 15
    metadata["rover_tilt_min"] = -50 if bottom_row is None else -80

    image_list = list()
    for path in source_paths:
        p = Path(path)
        if not p.exists():
            raise FileNotFoundError(f"{p} does not exist.")
            # in future, maybe do a db lookup on the VISID.

        image_list.append(imread(str(p)))

    pano_arr = np.hstack(image_list)

    if bottom_row is not None:
        if len(bottom_row) < len(source_paths):
            bottom_row += [
                "-",
            ] * (len(source_paths) - len(bottom_row))

        bottom_list = list()
        for b in bottom_row:
            if b == "-":
                bottom_list.append(np.zeros_like(image_list[0]))
            else:
                p = Path(b)
                if not p.exists():
                    raise FileNotFoundError(f"{p} does not exist.")
                metadata["source_products"].append([str(pds.VISID(p))])
                im = imread(str(p))
                if im.shape != image_list[0].shape:
                    im = resize(im, image_list[0].shape)
                bottom_list.append(im)
        bot_arr = np.hstack(bottom_list)
        pano_arr = np.vstack((pano_arr, bot_arr))

    pp = make_pano_record(metadata, pano_arr, outdir, thumb)

    if image_records and session is not None:
        bound_name = getattr(session.get_bind(), "name", None)
        purposes = set()
        start_times = []
        stop_times = []
        for ir in image_records:
            purposes.add(ir.verification_purpose)
            start_times.append(
                ir.start_time
                if bound_name != "sqlite"
                else ir.start_time.replace(tzinfo=timezone.utc)
            )
            stop_times.append(
                ir.stop_time
                if bound_name != "sqlite"
                else ir.stop_time.replace(tzinfo=timezone.utc)
            )

        purp = purposes.pop()
        if purp is not None:
            pp.purpose = purp

        pp.start_time = min(start_times)
        pp.stop_time = max(stop_times)

    if json:
        write_json(pp.asdict(), outdir)

    if session is not None:
        if image_records:
            to_add: List[Union[PanoRecord, JuncImagePano]] = [
                pp,
            ]
            for ir in image_records:
                a = JuncImagePano()
                a.image_record = ir
                a.pano_record = pp
                to_add.append(a)

            session.add_all(to_add)

        else:
            session.add(pp)

    return


def make_pano_record(
    metadata: dict,
    image: Union[ImageType, Path, None] = None,
    outdir: Path = Path.cwd(),
    thumb: Union[int, None] = None,
) -> PanoRecord:
    """
    Returns a PanoProduct created from the provided meta-data, and
    if *image* is a numpy array, it will also use write_tiff() to
    create a TIFF data product in *outdir* (defaults to current
    working directory).
    """
    pp = PanoRecord(**metadata)
    pid = pds.PanoID(pp.product_id)

    if image is not None:
        if isinstance(image, Path):
            tif_d = tif_info(image)
        else:
            desc = f"VIPER Panorama {pid}"

            logger.debug(desc)
            outpath = (outdir / str(pid)).with_suffix(".tif")

            imsave(
                str(outpath),
                image,
                check_contrast=False,
                description=desc,
                metadata=None,
            )
            tif_d = tif_info(outpath)

            if thumb is not None:
                # Scale down image to be no larger than thumb pixels
                max_dim = max(np.shape(image))
                if max_dim > thumb:
                    scale = max_dim / thumb
                    new_shape = tuple(int(x / scale) for x in np.shape(image))
                    image_th = rescale_intensity(
                        resize(image, new_shape), in_range="image", out_range="uint8"
                    )
                else:
                    image_th = image

                imsave(
                    outpath.stem + "_thumb.jpg",
                    image_th,
                    check_contrast=False,
                    description=desc,
                    metadata=None,
                )

        pp.update(tif_d)
    else:
        pp.update({"file_byte_offset": None, "file_data_type": None})

    pp.update(
        {
            "software_name": "vipersci",
            "software_version": vipersci.__version__,
            "software_type": "Python",
            "software_program_name": __name__,
        }
    )

    return pp



#!/usr/bin/env python
"""This module has tests for the image_statistics module."""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.
import argparse
import unittest

import numpy as np

from vipersci.vis import image_statistics


class TestParser(unittest.TestCase):
    def test_parser(self):
        parser = image_statistics.arg_parser()
        self.assertIsInstance(parser, argparse.ArgumentParser)


class TestCompute(unittest.TestCase):
    def setUp(self):
        self.image = np.array(
            [
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
                [0, 1, 1000, 3000, 3001, 3002, 3003, 3004, 3005, 3006, 4096],
            ],
            dtype=np.uint16,
        )

    def test_compute(self):
        # print(image_statistics.pprint(self.image))
        stats = image_statistics.compute(self.image)
        self.assertEqual(stats["blur"], 1.0)
        self.assertAlmostEqual(stats["mean"], 2374.36363636)
        self.assertAlmostEqual(stats["std"], 1310.43637928)
        self.assertEqual(stats["central_clear"], 13)



#!/usr/bin/env python
# coding: utf-8

"""Defines the VIS pan and tilt records tables using the SQLAlchemy ORM."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from sqlalchemy import (
    DateTime,
    Float,
)
from sqlalchemy.orm import mapped_column, validates

from vipersci.vis.db import Base
import vipersci.vis.db.validators as vld


class PanRecord(Base):
    """An object to represent rows in the pan_records table for VIS.  Each row
    represents a pan angle at a particular time."""

    # This class is derived from SQLAlchemy's orm.DeclarativeBase
    # which means that it has a variety of class properties that are
    # then swept up into properties on the instantiated object via
    # super().__init__().

    # The table represents many of these objects, so the __tablename__ is
    # plural while the class name is singular.
    __tablename__ = "pan_records"

    angle = mapped_column(
        Float,
        nullable=False,
        doc="Value in radians of the pan angle of the rover PTU.  Zero is forward (+x) "
        "relative to the rover frame.",
    )
    datetime = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        primary_key=True,
        doc="The datetime at which the pan angle was reported.",
    )

    @validates("datetime")
    def validate_datetime_asutc(self, key, value):
        return vld.validate_datetime_asutc(key, value)


class TiltRecord(Base):
    """An object to represent rows in the tilt_records table for VIS.  Each row
    represents a tilt angle at a particular time."""

    # This class is derived from SQLAlchemy's orm.DeclarativeBase
    # which means that it has a variety of class properties that are
    # then swept up into properties on the instantiated object via
    # super().__init__().

    # The table represents many of these objects, so the __tablename__ is
    # plural while the class name is singular.
    __tablename__ = "tilt_records"

    angle = mapped_column(
        Float,
        nullable=False,
        doc="Value in radians of the tilt angle of the rover PTU.  Zero is level "
        "relative to the rover frame.",
    )
    datetime = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        primary_key=True,
        doc="The datetime at which the pan angle was reported.",
    )

    @validates("datetime")
    def validate_datetime_asutc(self, key, value):
        return vld.validate_datetime_asutc(key, value)
